@@ -55,6 +55,7 @@
  * a GSettings instance for */
 #define SCHEMA_GENERAL         "org.gnome.desktop.wm.preferences"
 #define SCHEMA_METACITY        "org.gnome.metacity"
+#define SCHEMA_METACITY_THEME  "org.gnome.metacity.theme"
 #define SCHEMA_INTERFACE       "org.gnome.desktop.interface"
 
 #define SETTINGS(s) g_hash_table_lookup (settings_schemas, (s))
@@ -71,7 +72,8 @@ static GDesktopFocusMode focus_mode = G_DESKTOP_FOCUS_MODE_CLICK;
 static GDesktopFocusNewWindows focus_new_windows = G_DESKTOP_FOCUS_NEW_WINDOWS_SMART;
 static gboolean raise_on_click = TRUE;
 static gboolean attach_modal_dialogs = FALSE;
-static char* current_theme = NULL;
+static gchar *current_theme_name = NULL;
+static MetaThemeType current_theme_type = META_THEME_TYPE_GTK;
 static int num_workspaces = 4;
 static GDesktopTitlebarAction action_double_click_titlebar = G_DESKTOP_TITLEBAR_ACTION_TOGGLE_MAXIMIZE;
 static GDesktopTitlebarAction action_middle_click_titlebar = G_DESKTOP_TITLEBAR_ACTION_LOWER;
@@ -253,6 +255,13 @@ static MetaEnumPreference preferences_enum[] =
       },
       &placement_mode,
     },
+    {
+      { "type",
+        SCHEMA_METACITY_THEME,
+        META_PREF_THEME_TYPE,
+      },
+      &current_theme_type,
+    },
     { { NULL, 0, 0 }, NULL },
   };
 
@@ -375,14 +384,6 @@ static MetaStringPreference preferences_string[] =
       mouse_button_mods_handler,
       NULL,
     },
-    {
-      { "theme",
-        SCHEMA_METACITY,
-        META_PREF_THEME,
-      },
-      theme_name_handler,
-      NULL,
-    },
     {
       { KEY_TITLEBAR_FONT,
         SCHEMA_GENERAL,
@@ -407,6 +408,14 @@ static MetaStringPreference preferences_string[] =
       NULL,
       &cursor_theme,
     },
+    {
+      { "name",
+        SCHEMA_METACITY_THEME,
+        META_PREF_THEME_NAME,
+      },
+      theme_name_handler,
+      NULL,
+    },
     { { NULL, 0, 0 }, NULL },
   };
 
@@ -758,6 +767,35 @@ queue_changed (MetaPreference pref)
                                     changed_idle_handler, NULL, NULL);
 }
 
+static gboolean
+in_desktop (const gchar *name)
+{
+  const gchar *xdg_current_desktop;
+  gboolean result;
+  gchar **desktops;
+  gint i;
+
+  xdg_current_desktop = g_getenv ("XDG_CURRENT_DESKTOP");
+  if (xdg_current_desktop == NULL)
+    return FALSE;
+
+  result = FALSE;
+  desktops = g_strsplit (xdg_current_desktop, ":", -1);
+
+  for (i = 0; desktops[i] != NULL; i++)
+    {
+      if (g_strcmp0 (desktops[i], name) == 0)
+        {
+          result = TRUE;
+          break;
+        }
+    }
+
+  g_strfreev (desktops);
+
+  return result;
+}
+
 static void
 gtk_decoration_layout_changed (GtkSettings *settings,
                                GParamSpec  *pspec,
@@ -810,11 +848,15 @@ gtk_decoration_layout_changed (GtkSettings *settings,
   g_free (layout);
 }
 
+
 static void
 init_gtk_decoration_layout (void)
 {
   GtkSettings *settings;
 
+  if (!in_desktop ("GNOME-Flashback"))
+    return;
+
   settings = gtk_settings_get_default ();
 
   g_signal_connect (settings, "notify::gtk-decoration-layout",
@@ -823,6 +865,26 @@ init_gtk_decoration_layout (void)
   gtk_decoration_layout_changed (settings, NULL, NULL);
 }
 
+static void
+gtk_theme_name_changed (GtkSettings *settings,
+                        GParamSpec  *pspec,
+                        gpointer     user_data)
+{
+  if (current_theme_type == META_THEME_TYPE_GTK)
+    queue_changed (META_PREF_THEME_NAME);
+}
+
+static void
+init_gtk_theme_name (void)
+{
+  GtkSettings *settings;
+
+  settings = gtk_settings_get_default ();
+
+  g_signal_connect (settings, "notify::gtk-theme-name",
+                    G_CALLBACK (gtk_theme_name_changed), NULL);
+}
+
 /****************************************************************************/
 /* Initialisation.                                                          */
 /****************************************************************************/
@@ -843,6 +905,10 @@ meta_prefs_init (void)
   g_signal_connect (settings, "changed", G_CALLBACK (settings_changed), NULL);
   g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_METACITY), settings);
 
+  settings = g_settings_new (SCHEMA_METACITY_THEME);
+  g_signal_connect (settings, "changed", G_CALLBACK (settings_changed), NULL);
+  g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_METACITY_THEME), settings);
+
   /* Individual keys we watch outside of our schemas */
   settings = g_settings_new (SCHEMA_INTERFACE);
   g_signal_connect (settings, "changed::" KEY_GNOME_ACCESSIBILITY,
@@ -865,6 +931,7 @@ meta_prefs_init (void)
   init_workspace_names ();
 
   init_gtk_decoration_layout ();
+  init_gtk_theme_name ();
 }
 
 /****************************************************************************/
@@ -985,10 +1052,16 @@ meta_prefs_get_raise_on_click (void)
   return raise_on_click;
 }
 
-const char*
-meta_prefs_get_theme (void)
+const gchar *
+meta_prefs_get_theme_name (void)
 {
-  return current_theme;
+  return current_theme_name;
+}
+
+MetaThemeType
+meta_prefs_get_theme_type (void)
+{
+  return current_theme_type;
 }
 
 const char*
@@ -1057,17 +1130,12 @@ theme_name_handler (GVariant *value,
   *result = NULL; /* ignored */
   string_value = g_variant_get_string (value, NULL);
 
-  if (g_strcmp0 (current_theme, string_value) != 0)
+  if (g_strcmp0 (current_theme_name, string_value) != 0)
     {
-      if (current_theme)
-        g_free (current_theme);
+      g_free (current_theme_name);
+      current_theme_name = g_strdup (string_value);
 
-      if (!string_value || !*string_value)
-        current_theme = NULL;
-      else
-        current_theme = g_strdup (string_value);
-
-      queue_changed (META_PREF_THEME);
+      queue_changed (META_PREF_THEME_NAME);
     }
 
   return TRUE;
@@ -1115,7 +1183,7 @@ button_layout_equal (const MetaButtonLayout *a,
   int i;
 
   i = 0;
-  while (i < MAX_BUTTONS_PER_CORNER)
+  while (i < META_BUTTON_FUNCTION_LAST)
     {
       if (a->left_buttons[i] != b->left_buttons[i])
         return FALSE;
@@ -1131,238 +1199,14 @@ button_layout_equal (const MetaButtonLayout *a,
   return TRUE;
 }
 
-/*
- * This conversion cannot be handled by GSettings since
- * several values are stored in the same key (as a string).
- */
-static MetaButtonFunction
-button_function_from_string (const char *str)
-{
-  if (strcmp (str, "menu") == 0)
-    return META_BUTTON_FUNCTION_MENU;
-  else if (strcmp (str, "appmenu") == 0)
-    return META_BUTTON_FUNCTION_APPMENU;
-  else if (strcmp (str, "minimize") == 0)
-    return META_BUTTON_FUNCTION_MINIMIZE;
-  else if (strcmp (str, "maximize") == 0)
-    return META_BUTTON_FUNCTION_MAXIMIZE;
-  else if (strcmp (str, "close") == 0)
-    return META_BUTTON_FUNCTION_CLOSE;
-  else if (strcmp (str, "shade") == 0)
-    return META_BUTTON_FUNCTION_SHADE;
-  else if (strcmp (str, "above") == 0)
-    return META_BUTTON_FUNCTION_ABOVE;
-  else if (strcmp (str, "stick") == 0)
-    return META_BUTTON_FUNCTION_STICK;
-  else
-    /* don't know; give up */
-    return META_BUTTON_FUNCTION_LAST;
-}
-
-static MetaButtonFunction
-button_opposite_function (MetaButtonFunction ofwhat)
-{
-  switch (ofwhat)
-    {
-    case META_BUTTON_FUNCTION_SHADE:
-      return META_BUTTON_FUNCTION_UNSHADE;
-    case META_BUTTON_FUNCTION_UNSHADE:
-      return META_BUTTON_FUNCTION_SHADE;
-
-    case META_BUTTON_FUNCTION_ABOVE:
-      return META_BUTTON_FUNCTION_UNABOVE;
-    case META_BUTTON_FUNCTION_UNABOVE:
-      return META_BUTTON_FUNCTION_ABOVE;
-
-    case META_BUTTON_FUNCTION_STICK:
-      return META_BUTTON_FUNCTION_UNSTICK;
-    case META_BUTTON_FUNCTION_UNSTICK:
-      return META_BUTTON_FUNCTION_STICK;
-
-    case META_BUTTON_FUNCTION_MENU:
-    case META_BUTTON_FUNCTION_APPMENU:
-    case META_BUTTON_FUNCTION_MINIMIZE:
-    case META_BUTTON_FUNCTION_MAXIMIZE:
-    case META_BUTTON_FUNCTION_CLOSE:
-    case META_BUTTON_FUNCTION_LAST:
-      return META_BUTTON_FUNCTION_LAST;
-
-    default:
-      return META_BUTTON_FUNCTION_LAST;
-    }
-}
-
 static void
 update_button_layout (const gchar *string_value)
 {
+  gboolean invert;
   MetaButtonLayout new_layout;
-  char **sides = NULL;
-  int i;
 
-  /* We need to ignore unknown button functions, for
-   * compat with future versions
-   */
-
-  sides = g_strsplit (string_value, ":", 2);
-  i = 0;
-
-  if (sides != NULL && sides[0] != NULL)
-    {
-      char **buttons;
-      int b;
-      gboolean used[META_BUTTON_FUNCTION_LAST];
-
-      while (i < META_BUTTON_FUNCTION_LAST)
-        {
-          used[i] = FALSE;
-          new_layout.left_buttons_has_spacer[i] = FALSE;
-          i++;
-        }
-
-      buttons = g_strsplit (sides[0], ",", -1);
-      i = 0;
-      b = 0;
-      while (buttons[b] != NULL)
-        {
-          MetaButtonFunction f = button_function_from_string (buttons[b]);
-          if (i > 0 && strcmp("spacer", buttons[b]) == 0)
-            {
-              new_layout.left_buttons_has_spacer[i-1] = TRUE;
-              f = button_opposite_function (f);
-
-              if (f != META_BUTTON_FUNCTION_LAST)
-                {
-                  new_layout.left_buttons_has_spacer[i-2] = TRUE;
-                }
-            }
-          else
-            {
-              if (f != META_BUTTON_FUNCTION_LAST && !used[f])
-                {
-                  new_layout.left_buttons[i] = f;
-                  used[f] = TRUE;
-                  i++;
-
-                  f = button_opposite_function (f);
-
-                  if (f != META_BUTTON_FUNCTION_LAST)
-                      new_layout.left_buttons[i++] = f;
-
-                }
-              else
-                {
-                  meta_topic (META_DEBUG_PREFS, "Ignoring unknown or already-used button name \"%s\"\n",
-                              buttons[b]);
-                }
-            }
-
-          ++b;
-        }
-
-      g_strfreev (buttons);
-    }
-
-  /* Close the buttons list even if it's empty */
-  new_layout.left_buttons[i] = META_BUTTON_FUNCTION_LAST;
-  new_layout.left_buttons_has_spacer[i] = FALSE;
-
-
-  i = 0;
-
-  if (sides != NULL && sides[0] != NULL && sides[1] != NULL)
-    {
-      char **buttons;
-      int b;
-      gboolean used[META_BUTTON_FUNCTION_LAST];
-
-      while (i < META_BUTTON_FUNCTION_LAST)
-        {
-          used[i] = FALSE;
-          new_layout.right_buttons_has_spacer[i] = FALSE;
-          i++;
-        }
-
-      buttons = g_strsplit (sides[1], ",", -1);
-      i = 0;
-      b = 0;
-      while (buttons[b] != NULL)
-        {
-          MetaButtonFunction f = button_function_from_string (buttons[b]);
-          if (i > 0 && strcmp("spacer", buttons[b]) == 0)
-            {
-              new_layout.right_buttons_has_spacer[i-1] = TRUE;
-              f = button_opposite_function (f);
-              if (f != META_BUTTON_FUNCTION_LAST)
-                {
-                  new_layout.right_buttons_has_spacer[i-2] = TRUE;
-                }
-            }
-          else
-            {
-              if (f != META_BUTTON_FUNCTION_LAST && !used[f])
-                {
-                  new_layout.right_buttons[i] = f;
-                  used[f] = TRUE;
-                  i++;
-
-                  f = button_opposite_function (f);
-
-                  if (f != META_BUTTON_FUNCTION_LAST)
-                      new_layout.right_buttons[i++] = f;
-
-                }
-              else
-                {
-                  meta_topic (META_DEBUG_PREFS, "Ignoring unknown or already-used button name \"%s\"\n",
-                              buttons[b]);
-                }
-            }
-
-          ++b;
-        }
-
-      g_strfreev (buttons);
-    }
-
-  /* Close the buttons list even if it's empty */
-  new_layout.right_buttons[i] = META_BUTTON_FUNCTION_LAST;
-  new_layout.right_buttons_has_spacer[i] = FALSE;
-
-  if (sides)
-    g_strfreev (sides);
-
-  /* Invert the button layout for RTL languages */
-  if (meta_ui_get_direction() == META_UI_DIRECTION_RTL)
-  {
-    MetaButtonLayout rtl_layout;
-    int j;
-
-    for (i = 0; new_layout.left_buttons[i] != META_BUTTON_FUNCTION_LAST; i++);
-    for (j = 0; j < i; j++)
-      {
-        rtl_layout.right_buttons[j] = new_layout.left_buttons[i - j - 1];
-        if (j == 0)
-          rtl_layout.right_buttons_has_spacer[i - 1] = new_layout.left_buttons_has_spacer[i - j - 1];
-        else
-          rtl_layout.right_buttons_has_spacer[j - 1] = new_layout.left_buttons_has_spacer[i - j - 1];
-      }
-    rtl_layout.right_buttons[j] = META_BUTTON_FUNCTION_LAST;
-    rtl_layout.right_buttons_has_spacer[j] = FALSE;
-
-    for (i = 0; new_layout.right_buttons[i] != META_BUTTON_FUNCTION_LAST; i++);
-    for (j = 0; j < i; j++)
-      {
-        rtl_layout.left_buttons[j] = new_layout.right_buttons[i - j - 1];
-        if (j == 0)
-          rtl_layout.left_buttons_has_spacer[i - 1] = new_layout.right_buttons_has_spacer[i - j - 1];
-        else
-          rtl_layout.left_buttons_has_spacer[j - 1] = new_layout.right_buttons_has_spacer[i - j - 1];
-      }
-    rtl_layout.left_buttons[j] = META_BUTTON_FUNCTION_LAST;
-    rtl_layout.left_buttons_has_spacer[j] = FALSE;
-
-    new_layout = rtl_layout;
-  }
+  invert = meta_ui_get_direction() == META_UI_DIRECTION_RTL;
+  new_layout = meta_button_layout_new (string_value, invert);
 
   if (!button_layout_equal (&button_layout, &new_layout))
     {
@@ -1376,16 +1220,18 @@ button_layout_handler (GVariant *value,
                        gpointer *result,
                        gpointer  data)
 {
-#if 0
   const gchar *string_value;
 
+  *result = NULL; /* ignored */
+
+  if (in_desktop ("GNOME-Flashback"))
+    return TRUE;
+
   string_value = g_variant_get_string (value, NULL);
 
   if (string_value)
     update_button_layout (string_value);
-#endif
 
-  *result = NULL; /* ignored */
   return TRUE;
 }
 
@@ -1432,8 +1278,11 @@ meta_preference_to_string (MetaPreference pref)
     case META_PREF_RAISE_ON_CLICK:
       return "RAISE_ON_CLICK";
 
-    case META_PREF_THEME:
-      return "THEME";
+    case META_PREF_THEME_NAME:
+      return "THEME_NAME";
+
+    case META_PREF_THEME_TYPE:
+      return "THEME_TYPE";
 
     case META_PREF_TITLEBAR_FONT:
       return "TITLEBAR_FONT";
