@@ -30,9 +30,9 @@
 #include "core.h"
 #include "menu.h"
 #include "fixedtip.h"
-#include "theme.h"
 #include "prefs.h"
 #include "ui.h"
+#include "ui-private.h"
 
 #include <cairo-xlib.h>
 
@@ -57,8 +57,6 @@ static gboolean meta_frames_button_release_event  (GtkWidget           *widget,
                                                    GdkEventButton      *event);
 static gboolean meta_frames_motion_notify_event   (GtkWidget           *widget,
                                                    GdkEventMotion      *event);
-static gboolean meta_frames_destroy_event         (GtkWidget           *widget,
-                                                   GdkEventAny         *event);
 static gboolean meta_frames_draw                  (GtkWidget           *widget,
                                                    cairo_t             *cr);
 static gboolean meta_frames_enter_notify_event    (GtkWidget           *widget,
@@ -77,9 +75,6 @@ static void meta_frames_calc_geometry (MetaFrames        *frames,
                                        MetaUIFrame         *frame,
                                        MetaFrameGeometry *fgeom);
 
-static void meta_frames_ensure_layout (MetaFrames      *frames,
-                                       MetaUIFrame     *frame);
-
 static MetaUIFrame* meta_frames_lookup_window (MetaFrames *frames,
                                                Window      xwindow);
 
@@ -98,44 +93,39 @@ static void invalidate_all_caches (MetaFrames *frames);
 static void invalidate_whole_window (MetaFrames *frames,
                                      MetaUIFrame *frame);
 
-G_DEFINE_TYPE (MetaFrames, meta_frames, GTK_TYPE_WINDOW);
-
-
-static GObject *
-meta_frames_constructor (GType                  gtype,
-                         guint                  n_properties,
-                         GObjectConstructParam *properties)
+struct _MetaFrames
 {
-  GObject *object;
-  GObjectClass *gobject_class;
+  GtkWindow    parent;
 
-  gobject_class = G_OBJECT_CLASS (meta_frames_parent_class);
-  object = gobject_class->constructor (gtype, n_properties, properties);
+  GHashTable  *frames;
 
-  g_object_set (object,
-                "type", GTK_WINDOW_POPUP,
-                NULL);
+  guint        tooltip_timeout;
+  MetaUIFrame *last_motion_frame;
 
-  return object;
-}
+  gint         expose_delay_count;
+
+  gint         invalidate_cache_timeout_id;
+  GList       *invalidate_frames;
+  GHashTable  *cache;
+};
+
+G_DEFINE_TYPE (MetaFrames, meta_frames, GTK_TYPE_WINDOW)
 
 static void
-meta_frames_class_init (MetaFramesClass *class)
+meta_frames_class_init (MetaFramesClass *frames_class)
 {
-  GObjectClass   *gobject_class;
+  GObjectClass *object_class;
   GtkWidgetClass *widget_class;
 
-  gobject_class = G_OBJECT_CLASS (class);
-  widget_class = (GtkWidgetClass*) class;
+  object_class = G_OBJECT_CLASS (frames_class);
+  widget_class = GTK_WIDGET_CLASS (frames_class);
 
-  gobject_class->constructor = meta_frames_constructor;
-  gobject_class->finalize = meta_frames_finalize;
+  object_class->finalize = meta_frames_finalize;
 
   widget_class->destroy = meta_frames_destroy;
   widget_class->style_updated = meta_frames_style_updated;
 
   widget_class->draw = meta_frames_draw;
-  widget_class->destroy_event = meta_frames_destroy_event;
   widget_class->button_press_event = meta_frames_button_press_event;
   widget_class->button_release_event = meta_frames_button_release_event;
   widget_class->motion_notify_event = meta_frames_motion_notify_event;
@@ -175,50 +165,9 @@ prefs_changed_callback (MetaPreference pref,
     meta_frames_button_layout_changed (META_FRAMES (data));
 }
 
-static MetaStyleInfo *
-meta_frames_get_theme_variant (MetaFrames  *frames,
-                               const gchar *variant)
-{
-  MetaStyleInfo *style_info;
-
-  style_info = g_hash_table_lookup (frames->style_variants, variant);
-  if (style_info == NULL)
-    {
-      style_info = meta_theme_create_style_info (gtk_widget_get_screen (GTK_WIDGET (frames)), variant);
-      g_hash_table_insert (frames->style_variants, g_strdup (variant), style_info);
-    }
-
-  return style_info;
-}
-
-static void
-update_style_contexts (MetaFrames *frames)
-{
-  MetaStyleInfo *style_info;
-  GList *variants, *variant;
-  GdkScreen *screen;
-
-  screen = gtk_widget_get_screen (GTK_WIDGET (frames));
-
-  if (frames->normal_style)
-    meta_style_info_unref (frames->normal_style);
-  frames->normal_style = meta_theme_create_style_info (screen, NULL);
-
-  variants = g_hash_table_get_keys (frames->style_variants);
-  for (variant = variants; variant; variant = variants->next)
-    {
-      style_info = meta_theme_create_style_info (screen, (char *)variant->data);
-      g_hash_table_insert (frames->style_variants,
-                           g_strdup (variant->data), style_info);
-    }
-  g_list_free (variants);
-}
-
 static void
 meta_frames_init (MetaFrames *frames)
 {
-  frames->text_heights = g_hash_table_new (NULL, NULL);
-
   frames->frames = g_hash_table_new (unsigned_long_hash, unsigned_long_equal);
 
   frames->tooltip_timeout = 0;
@@ -229,10 +178,6 @@ meta_frames_init (MetaFrames *frames)
   frames->invalidate_frames = NULL;
   frames->cache = g_hash_table_new (g_direct_hash, g_direct_equal);
 
-  frames->style_variants = g_hash_table_new_full (g_str_hash, g_str_equal,
-                                                  g_free, (GDestroyNotify)meta_style_info_unref);
-  update_style_contexts (frames);
-
   meta_prefs_add_listener (prefs_changed_callback, frames);
 }
 
@@ -270,18 +215,6 @@ meta_frames_destroy (GtkWidget *widget)
     }
   g_slist_free (winlist);
 
-  if (frames->normal_style)
-    {
-      meta_style_info_unref (frames->normal_style);
-      frames->normal_style = NULL;
-    }
-
-  if (frames->style_variants)
-    {
-      g_hash_table_destroy (frames->style_variants);
-      frames->style_variants = NULL;
-    }
-
   GTK_WIDGET_CLASS (meta_frames_parent_class)->destroy (widget);
 }
 
@@ -294,8 +227,6 @@ meta_frames_finalize (GObject *object)
 
   meta_prefs_remove_listener (prefs_changed_callback, frames);
 
-  g_hash_table_destroy (frames->text_heights);
-
   invalidate_all_caches (frames);
   if (frames->invalidate_cache_timeout_id)
     g_source_remove (frames->invalidate_cache_timeout_id);
@@ -391,25 +322,18 @@ queue_recalc_func (gpointer key, gpointer value, gpointer data)
   invalidate_whole_window (frames, frame);
   meta_core_queue_frame_resize (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
                                 frame->xwindow);
-  if (frame->text_layout)
-    {
-      /* save title to recreate layout */
-      g_free (frame->title);
-
-      frame->title = g_strdup (pango_layout_get_text (frame->text_layout));
-
-      g_clear_object (&frame->text_layout);
-    }
 }
 
 static void
 meta_frames_font_changed (MetaFrames *frames)
 {
-  if (g_hash_table_size (frames->text_heights) > 0)
-    {
-      g_hash_table_destroy (frames->text_heights);
-      frames->text_heights = g_hash_table_new (NULL, NULL);
-    }
+  MetaTheme *theme;
+  const PangoFontDescription *titlebar_font;
+
+  theme = meta_ui_get_theme ();
+  titlebar_font = meta_prefs_get_titlebar_font ();
+
+  meta_theme_set_titlebar_font (theme, titlebar_font);
 
   /* Queue a draw/resize on all frames */
   g_hash_table_foreach (frames->frames,
@@ -452,12 +376,18 @@ static void
 meta_frames_style_updated (GtkWidget *widget)
 {
   MetaFrames *frames;
+  MetaTheme *theme;
+  gboolean compositing_manager;
 
   frames = META_FRAMES (widget);
 
-  meta_frames_font_changed (frames);
+  theme = meta_ui_get_theme ();
+  compositing_manager = meta_prefs_get_compositing_manager ();
 
-  update_style_contexts (frames);
+  meta_theme_set_composited (theme, compositing_manager);
+  meta_theme_invalidate (theme);
+
+  meta_frames_font_changed (frames);
 
   g_hash_table_foreach (frames->frames,
                         reattach_style_func, frames);
@@ -467,102 +397,6 @@ meta_frames_style_updated (GtkWidget *widget)
   GTK_WIDGET_CLASS (meta_frames_parent_class)->style_updated (widget);
 }
 
-static void
-meta_frames_ensure_layout (MetaFrames  *frames,
-                           MetaUIFrame *frame)
-{
-  GtkWidget *widget;
-  MetaFrameFlags flags;
-  MetaFrameType type;
-  MetaFrameStyle *style;
-
-  g_return_if_fail (gtk_widget_get_realized (GTK_WIDGET (frames)));
-
-  widget = GTK_WIDGET (frames);
-
-  meta_core_get (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), frame->xwindow,
-                 META_CORE_GET_FRAME_FLAGS, &flags,
-                 META_CORE_GET_FRAME_TYPE, &type,
-                 META_CORE_GET_END);
-
-  style = meta_theme_get_frame_style (meta_theme_get_current (),
-                                      type, flags);
-
-  if (style != frame->cache_style)
-    {
-      if (frame->text_layout)
-        {
-          /* save title to recreate layout */
-          g_free (frame->title);
-
-          frame->title = g_strdup (pango_layout_get_text (frame->text_layout));
-
-          g_clear_object (&frame->text_layout);
-        }
-    }
-
-  frame->cache_style = style;
-
-  if (frame->text_layout == NULL)
-    {
-      gpointer key, value;
-      MetaTheme *current;
-      PangoFontDescription *font_desc;
-      int size;
-
-      frame->text_layout = gtk_widget_create_pango_layout (widget, frame->title);
-
-      pango_layout_set_ellipsize (frame->text_layout, PANGO_ELLIPSIZE_END);
-      pango_layout_set_auto_dir (frame->text_layout, FALSE);
-      pango_layout_set_single_paragraph_mode (frame->text_layout, TRUE);
-
-      current = meta_theme_get_current ();
-
-      if (current->is_gtk_theme == FALSE)
-        {
-          double scale;
-
-          scale = meta_theme_get_title_scale (current, type, flags);
-
-          font_desc = meta_gtk_widget_get_font_desc (widget, scale,
-                                                     meta_prefs_get_titlebar_font ());
-        }
-      else
-        {
-          font_desc = meta_style_info_create_font_desc (frame->style_info);
-          meta_frame_style_apply_scale (style, font_desc);
-        }
-
-      size = pango_font_description_get_size (font_desc);
-
-      if (g_hash_table_lookup_extended (frames->text_heights,
-                                        GINT_TO_POINTER (size),
-                                        &key, &value))
-        {
-          frame->text_height = GPOINTER_TO_INT (value);
-        }
-      else
-        {
-          frame->text_height =
-            meta_pango_font_desc_get_text_height (font_desc,
-                                                  gtk_widget_get_pango_context (widget));
-
-          g_hash_table_replace (frames->text_heights,
-                                GINT_TO_POINTER (size),
-                                GINT_TO_POINTER (frame->text_height));
-        }
-
-      pango_layout_set_font_description (frame->text_layout,
-                                         font_desc);
-
-      pango_font_description_free (font_desc);
-
-      /* Save some RAM */
-      g_free (frame->title);
-      frame->title = NULL;
-    }
-}
-
 static void
 meta_frames_calc_geometry (MetaFrames        *frames,
                            MetaUIFrame       *frame,
@@ -580,18 +414,10 @@ meta_frames_calc_geometry (MetaFrames        *frames,
                  META_CORE_GET_FRAME_TYPE, &type,
                  META_CORE_GET_END);
 
-  meta_frames_ensure_layout (frames, frame);
-
   meta_prefs_get_button_layout (&button_layout);
 
-  meta_theme_calc_geometry (meta_theme_get_current (),
-                            frame->style_info,
-                            type,
-                            frame->text_height,
-                            flags,
-                            width, height,
-                            &button_layout,
-                            fgeom);
+  meta_theme_calc_geometry (meta_ui_get_theme (), frame->theme_variant,
+                            type, flags, width, height, &button_layout, fgeom);
 }
 
 MetaFrames*
@@ -620,8 +446,8 @@ meta_frames_new (int screen_number)
   return frames;
 }
 
-static const char *
-get_theme_variant_override (MetaFrames *frames)
+static const gchar *
+get_global_theme_variant (MetaFrames *frames)
 {
   GdkScreen *screen = gtk_widget_get_screen (GTK_WIDGET (frames));
   GtkSettings *settings = gtk_settings_get_for_screen (screen);
@@ -647,32 +473,22 @@ static void
 meta_frames_attach_style (MetaFrames  *frames,
                           MetaUIFrame *frame)
 {
-  gboolean has_frame;
-  char *variant = NULL;
-  const char *variant_override;
+  GdkDisplay *display;
+  gchar *variant;
 
-  if (frame->style_info != NULL)
-    meta_style_info_unref (frame->style_info);
+  display = gdk_display_get_default ();
+  variant = NULL;
 
-  variant_override = get_theme_variant_override (frames);
+  meta_core_get (GDK_DISPLAY_XDISPLAY (display), frame->xwindow,
+                 META_CORE_GET_THEME_VARIANT, &variant,
+                 META_CORE_GET_END);
 
-  if (variant_override)
-    variant = g_strdup (variant_override);
+  g_free (frame->theme_variant);
+
+  if (variant == NULL)
+    frame->theme_variant = g_strdup (get_global_theme_variant (frames));
   else
-    meta_core_get (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
-                   frame->xwindow,
-                   META_CORE_WINDOW_HAS_FRAME, &has_frame,
-                   META_CORE_GET_THEME_VARIANT, &variant,
-                   META_CORE_GET_END);
-
-  if (variant == NULL || strcmp(variant, "normal") == 0)
-    frame->style_info = meta_style_info_ref (frames->normal_style);
-  else
-    frame->style_info = meta_style_info_ref (meta_frames_get_theme_variant (frames,
-                                                                            variant));
-
-  if (variant_override)
-    g_free (variant);
+    frame->theme_variant = *variant != '\0' ? g_strdup (variant) : NULL;
 }
 
 void
@@ -690,14 +506,11 @@ meta_frames_manage_window (MetaFrames *frames,
 
   gdk_window_set_user_data (frame->window, frames);
 
-  frame->style_info = NULL;
+  frame->theme_variant = NULL;
 
   /* Don't set event mask here, it's in frame.c */
 
   frame->xwindow = xwindow;
-  frame->cache_style = NULL;
-  frame->text_layout = NULL;
-  frame->text_height = -1;
   frame->title = NULL;
   frame->expose_delayed = FALSE;
   frame->shape_applied = FALSE;
@@ -737,13 +550,10 @@ meta_frames_unmanage_window (MetaFrames *frames,
 
       g_hash_table_remove (frames->frames, &frame->xwindow);
 
-      meta_style_info_unref (frame->style_info);
+      g_free (frame->theme_variant);
 
       gdk_window_destroy (frame->window);
 
-      if (frame->text_layout)
-        g_object_unref (G_OBJECT (frame->text_layout));
-
       if (frame->title)
         g_free (frame->title);
 
@@ -779,19 +589,13 @@ meta_ui_frame_get_borders (MetaFrames       *frames,
 
   g_return_if_fail (type < META_FRAME_TYPE_LAST);
 
-  meta_frames_ensure_layout (frames, frame);
-
   /* We can't get the full geometry, because that depends on
    * the client window size and probably we're being called
    * by the core move/resize code to decide on the client
    * window size
    */
-  meta_theme_get_frame_borders (meta_theme_get_current (),
-                                frame->style_info,
-                                type,
-                                frame->text_height,
-                                flags,
-                                borders);
+  meta_theme_get_frame_borders (meta_ui_get_theme (), frame->theme_variant,
+                                type, flags, borders);
 }
 
 void
@@ -1201,8 +1005,6 @@ meta_frames_set_title (MetaFrames *frames,
   g_free (frame->title);
   frame->title = g_strdup (title);
 
-  g_clear_object (&frame->text_layout);
-
   invalidate_whole_window (frames, frame);
 }
 
@@ -2194,22 +1996,6 @@ meta_frames_motion_notify_event     (GtkWidget           *widget,
   return TRUE;
 }
 
-static gboolean
-meta_frames_destroy_event           (GtkWidget           *widget,
-                                     GdkEventAny         *event)
-{
-  MetaUIFrame *frame;
-  MetaFrames *frames;
-
-  frames = META_FRAMES (widget);
-
-  frame = meta_frames_lookup_window (frames, GDK_WINDOW_XID (event->window));
-  if (frame == NULL)
-    return FALSE;
-
-  return TRUE;
-}
-
 static void
 setup_bg_cr (cairo_t *cr, GdkWindow *window, int x_offset, int y_offset)
 {
@@ -2294,12 +2080,8 @@ populate_cache (MetaFrames *frames,
       return;
     }
 
-  meta_theme_get_frame_borders (meta_theme_get_current (),
-                                frame->style_info,
-                                frame_type,
-                                frame->text_height,
-                                frame_flags,
-                                &borders);
+  meta_theme_get_frame_borders (meta_ui_get_theme (), frame->theme_variant,
+                                frame_type, frame_flags, &borders);
 
   pixels = get_cache (frames, frame);
 
@@ -2315,28 +2097,30 @@ populate_cache (MetaFrames *frames,
    * size without any border added. */
 
   /* top */
-  pixels->piece[0].rect.x = 0;
-  pixels->piece[0].rect.y = 0;
-  pixels->piece[0].rect.width = width + borders.total.left + borders.total.right;
-  pixels->piece[0].rect.height = borders.total.top;
+  pixels->piece[0].rect.x = borders.invisible.left - borders.shadow.left;
+  pixels->piece[0].rect.y = borders.invisible.top - borders.shadow.top;
+  pixels->piece[0].rect.width = width + borders.visible.left + borders.shadow.left +
+                                borders.visible.right + borders.shadow.right;
+  pixels->piece[0].rect.height = borders.visible.top + borders.shadow.top;
 
   /* left */
-  pixels->piece[1].rect.x = 0;
+  pixels->piece[1].rect.x = borders.invisible.left - borders.shadow.left;
   pixels->piece[1].rect.y = borders.total.top;
   pixels->piece[1].rect.height = height;
-  pixels->piece[1].rect.width = borders.total.left;
+  pixels->piece[1].rect.width = borders.visible.left + borders.shadow.left;
 
   /* right */
   pixels->piece[2].rect.x = borders.total.left + width;
   pixels->piece[2].rect.y = borders.total.top;
-  pixels->piece[2].rect.width = borders.total.right;
+  pixels->piece[2].rect.width = borders.visible.right  + borders.shadow.right;
   pixels->piece[2].rect.height = height;
 
   /* bottom */
-  pixels->piece[3].rect.x = 0;
+  pixels->piece[3].rect.x = borders.invisible.left - borders.shadow.left;
   pixels->piece[3].rect.y = borders.total.top + height;
-  pixels->piece[3].rect.width = width + borders.total.left + borders.total.right;
-  pixels->piece[3].rect.height = borders.total.bottom;
+  pixels->piece[3].rect.width = width + borders.visible.left + borders.shadow.left +
+                                borders.visible.right + borders.shadow.right;
+  pixels->piece[3].rect.height = borders.visible.bottom + borders.shadow.bottom;
 
   for (i = 0; i < 4; i++)
     {
@@ -2355,6 +2139,39 @@ populate_cache (MetaFrames *frames,
       g_list_prepend (frames->invalidate_frames, frame);
 }
 
+static void
+clip_to_screen (cairo_region_t *region, MetaUIFrame *frame)
+{
+  cairo_rectangle_int_t frame_area;
+  cairo_rectangle_int_t screen_area = { 0, 0, 0, 0 };
+  cairo_region_t *tmp_region;
+
+  /* Chop off stuff outside the screen; this optimization
+   * is crucial to handle huge client windows,
+   * like "xterm -geometry 1000x1000"
+   */
+  meta_core_get (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), frame->xwindow,
+                 META_CORE_GET_FRAME_X, &frame_area.x,
+                 META_CORE_GET_FRAME_Y, &frame_area.y,
+                 META_CORE_GET_FRAME_WIDTH, &frame_area.width,
+                 META_CORE_GET_FRAME_HEIGHT, &frame_area.height,
+                 META_CORE_GET_SCREEN_WIDTH, &screen_area.width,
+                 META_CORE_GET_SCREEN_HEIGHT, &screen_area.height,
+                 META_CORE_GET_END);
+
+  cairo_region_translate (region, frame_area.x, frame_area.y);
+
+  tmp_region = cairo_region_create_rectangle (&frame_area);
+  cairo_region_intersect (region, tmp_region);
+  cairo_region_destroy (tmp_region);
+
+  tmp_region = cairo_region_create_rectangle (&screen_area);
+  cairo_region_intersect (region, tmp_region);
+  cairo_region_destroy (tmp_region);
+
+  cairo_region_translate (region, - frame_area.x, - frame_area.y);
+}
+
 static void
 subtract_client_area (cairo_region_t *region,
                       MetaUIFrame    *frame)
@@ -2374,10 +2191,9 @@ subtract_client_area (cairo_region_t *region,
                  META_CORE_GET_CLIENT_WIDTH, &area.width,
                  META_CORE_GET_CLIENT_HEIGHT, &area.height,
                  META_CORE_GET_END);
-  meta_theme_get_frame_borders (meta_theme_get_current (),
-                                frame->style_info,
-                                type, frame->text_height, flags,
-                                &borders);
+
+  meta_theme_get_frame_borders (meta_ui_get_theme (), frame->theme_variant,
+                                type, flags, &borders);
 
   area.x = borders.total.left;
   area.y = borders.total.top;
@@ -2427,12 +2243,7 @@ meta_frames_get_mask (MetaFrames *frames,
   if (frame == NULL)
     return;
 
-  cairo_push_group (cr);
-
   meta_frames_paint (frames, frame, cr);
-
-  cairo_pop_group_to_source (cr);
-  cairo_paint (cr);
 }
 
 /* XXX -- this is disgusting. Find a better approach here.
@@ -2484,6 +2295,7 @@ meta_frames_draw (GtkWidget *widget,
 
   cached_pixels_draw (pixels, cr, region);
 
+  clip_to_screen (region, frame);
   subtract_client_area (region, frame);
 
   n_areas = cairo_region_num_rectangles (region);
@@ -2641,23 +2453,11 @@ meta_frames_paint (MetaFrames   *frames,
                  META_CORE_GET_CLIENT_HEIGHT, &h,
                  META_CORE_GET_END);
 
-  meta_frames_ensure_layout (frames, frame);
-
   meta_prefs_get_button_layout (&button_layout);
 
-  meta_theme_draw_frame (meta_theme_get_current (),
-                         frame->style_info,
-                         cr,
-                         type,
-                         flags,
-                         w,
-                         h,
-                         frame->text_layout,
-                         frame->text_height,
-                         &button_layout,
-                         button_states,
-                         mini_icon,
-                         icon);
+  meta_theme_draw_frame (meta_ui_get_theme (), frame->theme_variant,
+                         cr, type, flags, w, h, frame->title,
+                         &button_layout, button_states, mini_icon, icon);
 }
 
 static gboolean
@@ -2786,11 +2586,17 @@ get_control (MetaFrames *frames,
   gboolean has_vert, has_horiz;
   gboolean has_north_resize;
   GdkRectangle client;
+  MetaFrameBorders borders;
 
   meta_frames_calc_geometry (frames, frame, &fgeom);
   get_client_rect (&fgeom, fgeom.width, fgeom.height, &client);
 
-  if (x < 0 || y < 0 || x > fgeom.width || y > fgeom.height)
+  borders = fgeom.borders;
+
+  if (x < borders.invisible.left - borders.resize.left ||
+      y < borders.invisible.top - borders.resize.top ||
+      x > fgeom.width - borders.invisible.right + borders.resize.right ||
+      y > fgeom.height - borders.invisible.bottom + borders.resize.bottom)
     return META_FRAME_CONTROL_NONE;
 
   if (POINT_IN_RECT (x, y, client))
@@ -2819,7 +2625,7 @@ get_control (MetaFrames *frames,
 
   if (POINT_IN_RECT (x, y, fgeom.title_rect))
     {
-      if (has_vert && y <= (fgeom.borders.invisible.top + TOP_RESIZE_HEIGHT) && has_north_resize)
+      if (has_vert && y <= TOP_RESIZE_HEIGHT && has_north_resize)
         return META_FRAME_CONTROL_RESIZE_N;
       else
         return META_FRAME_CONTROL_TITLE;
