@@ -19,17 +19,25 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#include "config.h"
+
 #include "prefs.h"
 #include "ui.h"
+#include "ui-private.h"
 #include "frames.h"
 #include "util.h"
 #include "menu.h"
 #include "core.h"
-#include "theme.h"
 
 #include <string.h>
 #include <stdlib.h>
 #include <cairo-xlib.h>
+#include <libmetacity/meta-theme.h>
+
+/**
+ * The current theme. (Themes are singleton.)
+ */
+static MetaTheme *meta_current_theme = NULL;
 
 static void meta_ui_accelerator_parse (const char      *accel,
                                        guint           *keysym,
@@ -63,6 +71,8 @@ meta_ui_init (int *argc, char ***argv)
   gdk_disable_multidevice ();
 #endif
 
+  gdk_set_allowed_backends ("x11");
+
   if (!gtk_init_check (argc, argv))
     meta_fatal ("Unable to open X display %s\n", XDisplayName (NULL));
 
@@ -97,7 +107,7 @@ static gboolean
 maybe_redirect_mouse_event (XEvent *xevent)
 {
   GdkDisplay *gdisplay;
-  GdkDeviceManager *gmanager;
+  GdkSeat *seat;
   GdkDevice *gdevice;
   MetaUI *ui;
   GdkEvent *gevent;
@@ -130,8 +140,8 @@ maybe_redirect_mouse_event (XEvent *xevent)
   if (gdk_window == NULL)
     return FALSE;
 
-  gmanager = gdk_display_get_device_manager (gdisplay);
-  gdevice = gdk_device_manager_get_client_pointer (gmanager);
+  seat = gdk_display_get_default_seat (gdisplay);
+  gdevice = gdk_seat_get_pointer (seat);
 
   /* If GDK already thinks it has a grab, we better let it see events; this
    * is the menu-navigation case and events need to get sent to the appropriate
@@ -670,85 +680,95 @@ meta_ui_theme_get_frame_borders (MetaUI           *ui,
                                  MetaFrameFlags    flags,
                                  MetaFrameBorders *borders)
 {
-  int text_height;
-  PangoContext *context;
-  const PangoFontDescription *font_desc;
-  PangoFontDescription *free_font_desc = NULL;
+  MetaTheme *theme;
+  const gchar *theme_variant;
 
-  if (meta_ui_have_a_theme ())
+  theme = meta_ui_get_theme ();
+  theme_variant = NULL;
+
+  meta_theme_get_frame_borders (theme, theme_variant, type, flags, borders);
+}
+
+MetaTheme *
+meta_ui_get_theme (void)
+{
+  return meta_current_theme;
+}
+
+static gchar *
+get_theme_name (MetaThemeType theme_type)
+{
+  gchar *theme_name;
+
+  if (theme_type == META_THEME_TYPE_METACITY)
     {
-      GdkDisplay *display = gdk_x11_lookup_xdisplay (ui->xdisplay);
-      GdkScreen *screen = gdk_display_get_screen (display, XScreenNumberOfScreen (ui->xscreen));
-      MetaStyleInfo *style_info = NULL;
-      MetaTheme *current;
-
-      style_info = meta_theme_create_style_info (screen, NULL);
-
-      context = gtk_widget_get_pango_context (GTK_WIDGET (ui->frames));
-      font_desc = meta_prefs_get_titlebar_font ();
-
-      current = meta_theme_get_current ();
-
-      if (!font_desc)
-        {
-          if (current->is_gtk_theme == FALSE)
-            {
-              GtkStyleContext *style = NULL;
-              GtkWidgetPath *widget_path;
-
-              style = gtk_style_context_new ();
-              gtk_style_context_set_screen (style, screen);
-              widget_path = gtk_widget_path_new ();
-              gtk_widget_path_append_type (widget_path, GTK_TYPE_WINDOW);
-              gtk_style_context_set_path (style, widget_path);
-              gtk_widget_path_free (widget_path);
-
-              gtk_style_context_get (style, GTK_STATE_FLAG_NORMAL,
-                                     "font", &free_font_desc,
-                                     NULL);
-
-              if (style != NULL)
-                g_object_unref (style);
-            }
-          else
-            {
-              free_font_desc = meta_style_info_create_font_desc (style_info);
-            }
-
-          font_desc = (const PangoFontDescription *) free_font_desc;
-        }
-
-      text_height = meta_pango_font_desc_get_text_height (font_desc, context);
-
-      meta_theme_get_frame_borders (current,
-                                    style_info,
-                                    type, text_height, flags,
-                                    borders);
-
-      if (free_font_desc)
-        pango_font_description_free (free_font_desc);
-
-      if (style_info != NULL)
-        meta_style_info_unref (style_info);
+      theme_name = g_strdup (meta_prefs_get_theme_name ());
     }
   else
     {
-      meta_frame_borders_clear (borders);
+      GtkSettings *settings;
+
+      settings = gtk_settings_get_default ();
+
+      g_object_get (settings, "gtk-theme-name", &theme_name, NULL);
     }
+
+  return theme_name;
+}
+
+static MetaTheme *
+load_theme (MetaThemeType  theme_type,
+            const gchar   *theme_name)
+{
+  MetaTheme *theme;
+  gboolean compositing_manager;
+  const PangoFontDescription *titlebar_font;
+  GError *error;
+
+  theme = meta_theme_new (theme_type);
+
+  compositing_manager = meta_prefs_get_compositing_manager ();
+  meta_theme_set_composited (theme, compositing_manager);
+
+  titlebar_font = meta_prefs_get_titlebar_font ();
+  meta_theme_set_titlebar_font (theme, titlebar_font);
+
+  error = NULL;
+  if (!meta_theme_load (theme, theme_name, &error))
+    {
+      g_warning ("%s", error->message);
+      g_error_free (error);
+
+      g_object_unref (theme);
+      return NULL;
+    }
+
+  return theme;
 }
 
 void
-meta_ui_set_current_theme (const char *name,
-                           gboolean    force_reload)
+meta_ui_reload_theme (void)
 {
-  meta_theme_set_current (name, force_reload);
-  meta_invalidate_default_icons ();
-}
+  MetaThemeType theme_type;
+  gchar *theme_name;
+  MetaTheme *theme;
 
-gboolean
-meta_ui_have_a_theme (void)
-{
-  return meta_theme_get_current () != NULL;
+  theme_type = meta_prefs_get_theme_type ();
+  theme_name = get_theme_name (theme_type);
+
+  theme = load_theme (theme_type, theme_name);
+  g_free (theme_name);
+
+  if (theme == NULL)
+    {
+      g_warning (_("Falling back to default GTK+ theme - Adwaita"));
+      theme = load_theme (META_THEME_TYPE_GTK, "Adwaita");
+    }
+
+  g_set_object (&meta_current_theme, theme);
+  g_assert (meta_current_theme);
+
+  meta_invalidate_default_icons ();
 }
 
 static void
